/*
wgt_vis_stretchy_01

james seo
www.lossless.net

draw an elastic grid of glove rect images,
in which the earlier glove images radiate away
from the current glove image.

this program is a visualization of data
from the white glove tracking project.
it requires input data files that are available from
www.whiteglovetracking.com
which must be placed in the proper data directory
specified by the dataPath variable.

this program will output 10059 tiff image files
if mustSaveOutputImages is set to true.
afterwards, you can save the image files as a quicktime movie
using the framerate of 29.97.

this program does not output the audio track.
you will have to add it yourself
after you have compiled the output movie.

based on wgt_viz_01_13

input data setup code and frame class
based on yellow_box_PROCESSING
by evan roth
www.ni9e.com
*/

import processing.video.*;

boolean canBeginDrawing = false;
boolean mustSaveOutputImages = false;

String appName = "wgt_vis_stretchy";
String appVer = "01";
int appW = 640;
int appH = appW*3/4;
int vidW = appW;
int vidH = appH;
String dataPath = "/Users/james/Documents/code/data/";
String imgsPath = dataPath+"billie-jean-640x480/";
PImage[] imgs;
int imgsTotalNum;
PImage curImg;
Runtime r;

frame FM;
frame FMS[];
int comma;
String temp_f, temp_x, temp_y, temp_w, temp_h;
int f, x, y, w, h;
float t, inc;
Movie inMovie;
PFont courier;
String[] lines;

int curFnum;
int initFnum;
int imgsRequestedNum;
int imgsCacheableNum = 1000;
int imgsCachedNum = 0;
int fnumDiff;

int curSrcX, curSrcY, curSrcW, curSrcH;
int curDstX, curDstY, curDstW, curDstH;
int minDstW = 4;
int minDstH = minDstW;

int[] maskArray;
int maskVal;
int maxMaskVal = 240;
float maskAdj;

PImage[] gloveImgs;
PImage curGloveImg;

int tempX, tempY;
int tempFnum;
int tempSrcX, tempSrcY, tempSrcW, tempSrcH;
int tempBoundLeft, tempBoundRight, tempBoundUp, tempBoundDown;
int tempSideNum;

float easing = 0.1;
float dx, dy, dw, dh;
int easedX, easedY, easedW, easedH;

/***************************************************************
* SETUP
***************************************************************/

void setup(){
  size(appW, appH);
  stroke(0);
  fill(0);
  
  courier = loadFont(dataPath+"CourierNewPSMT-14.vlw");
  textFont(courier, 16);
  
  lines = loadStrings(dataPath+"wgt_data_v1.txt");
  println("number of lines in data file: " + lines.length);

  curFnum = 0;
  
  println("setting up input data");
  setupInputData();
  println("finished setting up input data");
  println("setting up input images");
  setupInputImages();
  println("finished setting up input images");
  
  curImg = createImage(vidW, vidH, ARGB);
  curSrcW = -1;
  curSrcH = -1;
}

void setupInputData() {
  t = 0.0f;
  inc = 1.0f / 30.0f;

  FMS = new frame[lines.length];

  for (int i=0; i < lines.length; i++) {
    comma = 0;
    temp_f = "";
    temp_x = "";
    temp_y = "";
    temp_w = "";
    temp_h = "";

    for (int j=0; j < lines[i].length(); j++) {
      char c = lines[i].charAt(j);    //this is the char we are looking at
      if(c == ','){                   //which comma are we on
        comma++;
      }
      if(comma == 0){                //break into multiple strings
        temp_f += c;
      }else if(comma == 1){
        if(c != ','){
          temp_x += c;
        }
      }else if(comma == 2){
        if(c != ','){
          temp_y += c;
        }
      }else if(comma == 3){
        if(c != ','){
          temp_h += c;
        }
      }else if(comma == 4){
        if(c != ','){
          temp_w += c;
        }
      }
    }

    f = int(temp_f);
    x = int(temp_x);
    y = int(temp_y);
    w = int(temp_w);
    h = int(temp_h);
    t += inc;

    if (x >= 640) {
      x = 639;
    }
    if (y >= 480) {
      y = 479;
    }
    if (x+w/2 >= 640) {
      w = 639-(x-w/2);
      x = 639-w/2;
    }
    if (y+h/2 >= 480) {
      h = 479-(y-h/2);
      y = 479-h/2;
    }
    
    FM = new frame(f, x, y, w, h, t);
    FMS[i] = FM;
    //FMS[i].print();
  }
}

void setupInputImages() {
  imgsTotalNum = lines.length;
  
  imgs = new PImage[imgsTotalNum];
  
  gloveImgs = new PImage[imgsTotalNum];
}

/***************************************************************
* DRAWING
***************************************************************/

void draw(){
  background(0);
  if (!canBeginDrawing) {
    initFnum = 0;
    imgsRequestedNum = imgsTotalNum-initFnum;
    curFnum = initFnum;
    canBeginDrawing = true;
  } else {
    println("current frame: " + curFnum);
    
    curImg = loadImage(imgsPath+"mj_bet_HD"+nf(curFnum+1,5)+".jpg");
    imgsCachedNum++;
    
    if (FMS[curFnum].x > 0) {
      curSrcX = FMS[curFnum].x - (FMS[curFnum].w / 2);
      curSrcY = FMS[curFnum].y - (FMS[curFnum].h / 2);
      curSrcW = FMS[curFnum].w;
      curSrcH = FMS[curFnum].h;
    }
    updateEasingParams();
    curDstX = easedX;
    curDstY = easedY;
    curDstW = easedW;
    curDstH = easedH;
    
    if (curDstW >= minDstW && curDstH >= minDstH) {
      if (FMS[curFnum].x > 0) {
        curGloveImg = curImg.get(curSrcX,curSrcY,curSrcW,curSrcH);
        createMaskArray(curGloveImg.width,curGloveImg.height);
        applyMask(curGloveImg,curFnum,curFnum);
        drawGloveImg(curGloveImg,curDstX,curDstY,curDstW,curDstH);
      } else {
        createMaskArray(curGloveImg.width,curGloveImg.height);
      }
      
      tempFnum = curFnum-1;
      while (tempFnum >= 0) {
        fnumDiff = abs(curFnum-tempFnum);
        updateTempBounds(curDstX,curDstY,curDstW,curDstH,fnumDiff);
        if (tempBoundLeft <= -curDstW && tempBoundRight >= appW
          && tempBoundUp <= -curDstH && tempBoundDown >= appH) {
          break;
        }
        tempSideNum = 1+2*(fnumDiff);
        if (FMS[tempFnum].x > 0) {
          tempSrcX = FMS[tempFnum].x - (FMS[tempFnum].w / 2);
          tempSrcY = FMS[tempFnum].y - (FMS[tempFnum].h / 2);
          tempSrcW = FMS[tempFnum].w;
          tempSrcH = FMS[tempFnum].h;
          if (imgs[tempFnum] == null) {
            imgs[tempFnum] = loadImage(imgsPath+"mj_bet_HD"+nf(tempFnum+1,5)+".jpg");
            imgsCachedNum++;
          }
          gloveImgs[tempFnum] = imgs[tempFnum].get(tempSrcX,tempSrcY,tempSrcW,tempSrcH);
          resizeMaskArray(gloveImgs[tempFnum].width,gloveImgs[tempFnum].height);
          applyMask(gloveImgs[tempFnum],curFnum,tempFnum);
          
          tempY = curDstY - (curDstH * fnumDiff);
          for (int i = 0; i < tempSideNum; i++) {
            tempX = curDstX + (curDstW * (i-fnumDiff));
            if ((tempX > -curDstW && tempX < appW)
              || (tempY > -curDstH && tempY < appH)) {
              drawGloveImg(gloveImgs[tempFnum],tempX,tempY,curDstW,curDstH);
            }
          }
          tempY = curDstY + (curDstH * fnumDiff);
          for (int i = 0; i < tempSideNum; i++) {
            tempX = curDstX + (curDstW * (i-fnumDiff));
            if ((tempX > -curDstW && tempX < appW)
              || (tempY > -curDstH && tempY < appH)) {
              drawGloveImg(gloveImgs[tempFnum],tempX,tempY,curDstW,curDstH);
            }
          }
          tempX = curDstX - (curDstW * fnumDiff);
          for (int i = 1; i < tempSideNum-1; i++) {
            tempY = curDstY + (curDstH * (i-fnumDiff));
            if ((tempX > -curDstW && tempX < appW)
              || (tempY > -curDstH && tempY < appH)) {
              drawGloveImg(gloveImgs[tempFnum],tempX,tempY,curDstW,curDstH);
            }
          }
          tempX = curDstX + (curDstW * fnumDiff);
          for (int i = 1; i < tempSideNum-1; i++) {
            tempY = curDstY + (curDstH * (i-fnumDiff));
            if ((tempX > -curDstW && tempX < appW)
              || (tempY > -curDstH && tempY < appH)) {
              drawGloveImg(gloveImgs[tempFnum],tempX,tempY,curDstW,curDstH);
            }
          }
        }
        tempFnum--;
      }
      
      deleteMaskArray();
    }
    
    if (mustSaveOutputImages) {
      save(appName + "_" + curFnum +".tif");
    }
  
    curFnum++;
    if (curFnum >= initFnum+imgsRequestedNum) {
      println(curFnum);
      println(initFnum);
      println(imgsRequestedNum);
      println("done!");
      exit();
    }
  
    curImg = null;
  
    if (imgsCachedNum > imgsCacheableNum) {
      for (int i = 0; i < imgs.length; i++) {
        imgs[i] = null;
      }
      for (int i = 0; i < gloveImgs.length; i++) {
        gloveImgs[i] = null;
      }
      imgsCachedNum = 0;
    }
  
    r = Runtime.getRuntime();
    r.gc();
  }
}

void drawGloveImg(PImage img, int x, int y, int w, int h) {
  image(img,x,y,w,h);
}

void updateTempBounds(int x, int y, int w, int h, int fd) {
  tempBoundLeft = x - (w * fd);
  tempBoundRight = x + (w * fd);
  tempBoundUp = y - (h * fd);
  tempBoundDown = y + (h * fd);
}

/***************************************************************
* IMAGE MASK
***************************************************************/

void createMaskArray(int w, int h) {
  maskArray = new int[w*h];
}

void deleteMaskArray() {
  maskArray = null;
}

void resizeMaskArray(int w, int h) {
  if (maskArray.length < w*h) {
    maskArray = expand(maskArray,w*h);
  } else if (maskArray.length > w*h) {
    maskArray = subset(maskArray,0,w*h);
  }
}

void fillMaskArray(int val) {
  for (int i = 0; i < maskArray.length; i++) {
    maskArray[i] = val;
  }
}

void applyMask(PImage img, int reffn, int tempfn) {
  if (img.pixels.length == maskArray.length) {
    maskVal = getMaskVal(reffn, tempfn);
    fillMaskArray(maskVal);
    img.mask(maskArray);
    img.updatePixels();
  } else {
    println("error: image and mask sizes are unequal");
  }
}

int getMaskVal(int reffn, int tempfn) {
  if (reffn == tempfn) {
    return 255;
  } else {
    maskAdj = 1-(abs(reffn-tempfn)/((float) 180));
    if (maskAdj <= 0.0001) {
      maskAdj = 0.0001;
    } else if (maskAdj >= 1.0) {
      maskAdj = 0.9999;
    }
    maskAdj = tan((6*PI/32)+(9*PI/32)*maskAdj)/tan(15*PI/32);
    if (maskAdj >= 0.9999) {
      maskAdj = 0.9999;
    }
    return (int) (maxMaskVal*maskAdj);
  }
}

/***************************************************************
* EASING
***************************************************************/

void updateEasingParams() {
  if (curFnum == initFnum) {
    easedX = curSrcX;
    easedY = curSrcY;
    easedW = curSrcW;
    easedH = curSrcH;
  } else {
    dx = curSrcX-easedX;
    if(abs(dx) > 1) {
      easedX += dx * easing;
    }
    dy = curSrcY-easedY;
    if(abs(dy) > 1) {
      easedY += dy * easing;
    }
    dw = curSrcW-easedW;
    if(abs(dw) > 1) {
      easedW += dw * easing;
    }
    dh = curSrcH-easedH;
    if(abs(dh) > 1) {
      easedH += dh * easing;
    }
  }
}